import pandas as pd
from langassist_model_engine.nlp_utils.tokenize import tokenize
import numpy as np
from scipy import sparse

import logging

logger = logging.getLogger(__name__)

'''
Features:
    - Num tokens
    - Average word length
    - log of the freq. of each character n-gram, normalized wrt text length.
        Smoothing term added to avoid 0 values for n-grams with 1 freq.
    - Word n-grams (presence or absence)
'''

####################################
###     Independent Features     ###
####################################
def num_tokens(text):
    return len(tokenize(text))

def average_word_len(text):
    '''
    Get the sum of the word lengths and divide by the number of words
    :param text:
    :return:
    '''
    word_list = tokenize(text)
    num_words = len(word_list)

    return sum([len(s) for s in word_list]) / num_words

def build_sentence_features(text):
    num_tok = num_tokens(text)
    avg_len = average_word_len(text)

    return (num_tok, avg_len)

##########################################
###     Total Feature Construction     ###
##########################################

def create_series(features):
    return pd.Series({'len': features[0],
                      'avg_word_len': features[1]})

def sentence_features_sparse(s_data, feature_builder):
    vf = np.vectorize(lambda s: feature_builder.featurize(s))
    s_data = s_data.ravel()
    s_features = vf(s_data)

    return sparse.vstack(s_features)



import os

DATA_ROOT = "data/NLI/data/text/"

NLI_DEV = "index-dev.csv"

ENVIRONMENT = "dev"

DATA_TOKENIZED = "tokenized/"
DATA_ORIGINAL = "original/"

DATA_FORMAT = DATA_TOKENIZED

PROJECT_ROOT = os.getcwd()
LOCAL_DUMP = PROJECT_ROOT + "/dumps/"
LOGIREG_PICKLE = PROJECT_ROOT + "/dumps/logireg.pkl"

CHAR_N_RANGE = (4, 8)
WORD_N_RANGE = (1, 3)



import argparse
import logging

from sklearn import preprocessing
from sklearn.model_selection import train_test_split

from langassist_model_engine.native_lang_identification.data import prep_data
from langassist_model_engine.native_lang_identification.model.base import BaseModel
from langassist_model_engine.native_lang_identification.model import logireg
from langassist_model_engine.nlp_utils.features.builders.get_builder import get_builder

logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s:\t %(message)s', level=logging.DEBUG)
logger = logging.getLogger(__name__)

def parse_args():
    '''
        Argument Parsing
    '''
    parser = argparse.ArgumentParser(description='''For training models''')
    parser.add_argument('--model', '-m',
                        help="The model to train. Currently only 'nli'",
                        type=str,
                        default='nli',
                        metavar="MODEL")
    parser.add_argument('--feature-builder', '-fb',
                        help="Feature Builder to use (vocab, hash)",
                        type=str,
                        default='hash',
                        metavar='FEAT-BUILDER')
    parser.add_argument('--num-samples',
                        help='The number of samples to use. 0 is default and means "use all"',
                        type=int,
                        default=0,
                        metavar="NUM_SAMPLES")
    parser.add_argument('--n-features',
                        help="Maximum number of features for (vocab model only)",
                        type=int,
                        default=0,
                        metavar="N_FEATURES")
    parser.add_argument('--produce-graphs', '-g',
                        action='store_true',
                        help="Specify whether or not to produce graphs",
                        default=False)
    parser.add_argument('--latest', '-l',
                        action='store_true',
                        help="Specify whether or not this model will be the current latest version",
                        default=False)
    parser.add_argument('--test', '-t',
                        action='store_true',
                        help="Specify whether or not to test the model using a test/train split",
                        default=False)
    parser.add_argument('--prod', '-p',
                        action='store_true',
                        help="Specify whether or not the model is intended for production use",
                        default=False)
    parser.add_argument('--word-ngram-range', '-wnr',
                        help='Specify the range for word n-grams',
                        nargs=2,
                        default=(1, 3),
                        metavar="W_NGRAM")
    parser.add_argument('--char-ngram-range', '-cnr',
                        help='Specify the range for char n-grams',
                        nargs=2,
                        default=(3, 7),
                        metavar="C_NGRAM")
    parser.add_argument('--to-pickle', '-tp',
                        help="Whether or not to save the model to a pickle",
                        action='store_true',
                        default=True)
    parser.add_argument('--from-pickle', '-fp',
                        help="Whether or not to load the model from a pickle",
                        type=str,
                        default="")

    return parser.parse_args()

def _ngr_to_int(ngr):
    '''
        Turn a tuple of str into a tuple of int.
        For cmd line parsing
    :param ngr:
    :return:
    '''
    # ngr means n-gram-range
    # TODO change ngr to something that will make sense when I read this again in 2 months
    return (int(ngr[0]), int(ngr[1]))

def train(opts):
    '''
        Train the required model, parameterised by the command line options
    :param opts:
    :return:
    '''
    logger.info("Preparing training data...")
    doc_data, s_data = prep_data.get_training_data(opts.num_samples)
    # Sentence labels
    Y_train = s_data['l1']
    logger.debug("%s unique labels" % Y_train.nunique())
    s_data = s_data.drop('l1', axis=1)

    min_max_scaler = preprocessing.MaxAbsScaler()  # Used to normalise data between 0 and 1

    if(opts.from_pickle is not ""):
        dir, key = opts.from_pickle.split('/')
        model = BaseModel.load(dir, key)
        feature_builder = model.feature_builder
    else:
        logger.info("Constructing feature builder and scaler...")
        # Get the requested feature builder class and then instantiate it
        fb_cls = get_builder(opts.feature_builder)
        feature_builder = fb_cls(_ngr_to_int(opts.word_ngram_range), _ngr_to_int(opts.word_ngram_range))
        model = logireg.LogiReg(feature_builder)
        # Encode string labels as integers
        model.fit_label_encoder(Y_train)

    logger.info("Featurizing data...")
    X_train = feature_builder.featurize_all(s_data.values.ravel())

    if(opts.test):
        X_train, x_test, Y_train, y_test = train_test_split(X_train, Y_train, test_size=0.9, random_state=42)

        X_train = min_max_scaler.fit_transform(X_train)
        x_test = min_max_scaler.fit_transform(x_test)

        model.train(X_train, Y_train)
        model.test(x_test, y_test)
    else:
        X_train = min_max_scaler.fit_transform(X_train)
        model.train(X_train, Y_train)

    if(opts.to_pickle):
        model.save(opts.prod, latest=opts.latest)


if __name__ == "__main__":
    train(parse_args())


def build_n_grams(items, n):
    """
    This function simply splits the items into n-tuples representing n-grams.
    The * operator splits a list into a series of arguments passed to a function.
    So effectively, this function calls:
        return zip(items[1:], items[2:], ..., items[n:])

    Pass a `str` for char n-grams, or a `list` of tokens for word n-grams
    :param items:
    :param n:
    :return a list of n-tuples representing n-grams of the tokens:
    """
    return list(zip(*[items[i:] for i in range(n)]))


import logging
from time import time

import numpy as np
from scipy import sparse
from sklearn.feature_extraction.text import HashingVectorizer

from langassist_model_engine.nlp_utils.features import sentence_features
from langassist_model_engine.nlp_utils.features.builders.base import BaseFeaturizer

logger = logging.getLogger(__name__)

class HashFeaturizer(BaseFeaturizer):
    '''
    This class will turn a piece of text into a *vocabulary-free* set of features

    `data_df` is the pandas DataFrame which will be used to construct the vocabulary.
    It should have the form

                document            l1
                "some text"         KOR
                "more text"         ESP
                ...                 ...

    The text in the `document` column is used to construct a vocabulary of ngram features for
    both characters and words.
    The vocabulary will be used to construct a vector whose length is the sum of the lengths of
    all the ngram vocabularies for that type {char, word}
    '''
    def __init__(self, word_range, char_range):
        BaseFeaturizer.__init__(self)
        self.word_range = word_range
        self.char_range = char_range
        self.word_pres_vec = HashingVectorizer(ngram_range=self.word_range, binary=True)
        self.char_count_vec = HashingVectorizer(ngram_range=self.char_range, analyzer='char', binary=False)


    def featurize(self, text):
        s_feats = sentence_features.build_sentence_features(text)
        char_counts = self.count_chars(text)
        word_presence = self.word_presence(text)

        ngram_features = np.concatenate((word_presence, char_counts))
        total_features = np.insert(ngram_features, 0, list(s_feats))

        return sparse.csr_matrix(total_features)

    def featurize_all(self, samples):
        '''
            Turn all samples into features.
            Samples is expected as a pandas DataFrame
        :param samples:
        :return:
        '''
        logger.info("Building Features...")
        t0 = time()

        vf = np.vectorize(lambda s: self.featurize(s))

        s_features = vf(samples)
        logger.debug("Done in %0.3fs" % (time() - t0))

        return sparse.vstack(s_features)


    def count_chars(self, text):
        '''
        Generate the ngram count vector, which is a vector of real values indicating
        the count of a char ngram in the sample text
        :param data:
        :return:
        '''
        return self.char_count_vec.fit_transform([text]).toarray()[0]


    def word_presence(self, text):
        '''
        Generate the word_presence_vector, which is a V long vector of binary values indicating
        whether or not a word is present
        :param text_col:
        :return:
        '''
        return self.word_pres_vec.fit_transform([text]).toarray()[0]


from sklearn.linear_model import LogisticRegression

from langassist_model_engine.native_lang_identification.model.base import BaseModel


class SS_Model(BaseModel):
    def __init__(self, train_x, train_y):
        BaseModel.__init__(self)
        self.model = LogisticRegression()

    def train(self, train_x, train_y):
        pass

    def test(self, test_x, test_y):
        pass

    def predict(self, predict_x):
        pass







from datetime import datetime
import logging
import os

import s3io
from sklearn.externals import joblib

from langassist_model_engine.constants import LOCAL_DUMP

logger = logging.getLogger(__name__)

class BaseModel():
    def __init__(self, feature_builder):
        self.feature_builder = feature_builder

    def train(self, x_train, y_train):
        raise NotImplementedError

    def predict(self, predict_x):
        raise NotImplementedError

    def test(self, x_test, y_test):
        raise NotImplementedError

    @staticmethod
    def _dump_dir():
        return "nli"

    def _archive_s3_models(self):
        pass

    def _archive_local_models(self):
        pass

    def _save_to_s3(self, latest=False):
        '''
        Save the model as a pickle in S3
        :return:
        '''
        credentials = dict(
            aws_access_key_id=os.environ['LANGASSIST_AWS_ACCESS_KEY'],
            aws_secret_access_key=os.environ['LANGASSIST_AWS_SECRET_KEY'],
        )
        bucket = os.environ['MODEL_BUCKET']
        key = self._pickle_name(latest)
        dump_dir = self._dump_dir()
        compress = ('gzip', 3)
        loc = 's3://{0}/{1}/{2}'.format(bucket, dump_dir, key)

        with s3io.open(
                loc,
                mode='w',
                **credentials
        ) as s3_file:
            joblib.dump(self, s3_file, compress=compress)

        logger.info("Saved model %s externally to %s" % (self.__str__(), loc))

    def _save_locally(self, latest=False):
        '''
        Save the model as a pickle in the local filesystem
        :return:
        '''

        dump_dir = self._dump_dir()
        loc = LOCAL_DUMP + dump_dir + '/' + self._pickle_name(latest)
        joblib.dump(self, loc)
        logger.info("Saved model %s locally to %s" % (self.__str__(), loc))

    def save(self, prod, latest=False):
        if(prod is True):
            self._archive_s3_models()
            self._save_to_s3(latest=latest)
        else:
            self._archive_local_models()
            self._save_locally(latest)

    @staticmethod
    def load(dir, key):
        credentials = dict(
            aws_access_key_id=os.environ['LANGASSIST_AWS_ACCESS_KEY'],
            aws_secret_access_key=os.environ['LANGASSIST_AWS_SECRET_KEY'],
        )
        bucket = os.environ['MODEL_BUCKET']

        with s3io.open(
                's3://{0}/{1}/{2}'.format(bucket, dir, key),
                mode='w',
                **credentials) as s3_file:
            return joblib.load(s3_file)

    def _pickle_name(self, latest=False):
        now = datetime.now()
        return "%s-%s_%d.%d.%d.%d.%d%s.pkl" % (self.__class__.__name__.lower(),
                                               self.feature_builder.__class__.__name__.lower(),
                                               now.year,
                                               now.month,
                                               now.day,
                                               now.hour,
                                               now.minute,
                                               "_latest" if latest else "")

    def __str__(self):
        return "[%s\t%s]" % (self.__class__.__name__, self.feature_builder.__class__.__name__)


import nltk
import string
from nltk.corpus import stopwords

sentence_tokenizer = nltk.data.load('tokenizers/punkt/english.pickle')

def to_sentences(document):
    '''
    Take a string as input and splits the string into sentences
    :param document:
    :return a list of strings, where each item is a sentence of the document:
    '''
    split_sentences = sentence_tokenizer.tokenize(document)

    return split_sentences

def tokenize(text):
    '''
    return the piece of text as a list of strings, where each string is a word
    :param text:
    :return:
    '''
    return nltk.word_tokenize(text)

def untokenize(tokens):
    '''
    Turn a list of tokens into a fully formed sentence
    Attempts to preserve things like "can't"
    :param tokens:
    :return:
    '''
    return "".join([" " + i if not i.startswith("'")
                               and i not in string.punctuation
                            else i
                    for i in tokens]).strip()

def remove_stops(tokens):
    '''
    Stops words are common word without much semantic value
    e.g. "of", "and", "while", etc.
    :param tokens:
    :return:
    '''
    stop_words = list(stopwords.words('english'))
    return [t for t in tokens if t not in stop_words]

def words_only(total_tokens):
    '''
    Filter the tokens to remove tokens which are not English words
    :param total_tokens:
    :return:
    '''
    raise NotImplementedError


import pandas as pd

from langassist_model_engine.constants import DATA_ROOT, NLI_DEV, DATA_FORMAT
from langassist_model_engine.nlp_utils.tokenize import *

####################
###  Load Files  ###
####################

def load_csv(file_name, num_samples):
    if(num_samples == 0):
        return pd.read_csv(file_name, delimiter=',')
    return pd.read_csv(file_name, delimiter=',').sample(num_samples)

def load_file(file):
    with open(file) as f:
        return f.read()

##########################
###  Build File Names  ###
##########################

def build_index_file_name(index_name):
    return ''.join((DATA_ROOT, index_name))

def build_data_file_name(data_name):
    return ''.join((DATA_ROOT, "responses/", DATA_FORMAT, data_name))

###########################
###  Build Data Frames  ###
###########################

def load_total_data(index_file, num_samples):
    ind_df = load_csv(build_index_file_name(index_file), num_samples)

    ind_df['Filename'] = ind_df['Filename'].apply(lambda x: load_file(build_data_file_name(x)))


    return ind_df

def get_sentence_pairs(doc_pairs):
    sent_pairs = pd.DataFrame(columns=['sentence', 'l1'])
    for index, row in doc_pairs.iterrows():
        l1 = row['l1']
        sentences = to_sentences(row['document'])
        for s in sentences:
            sent_pairs = sent_pairs.append({'sentence': s, 'l1': l1}, ignore_index=True)

    return sent_pairs

###############################
###  Prepare Training Data  ###
###############################

def get_training_data(num_samples):
    document_pairs = load_total_data(NLI_DEV, num_samples)
    document_pairs = document_pairs.rename(columns={'Filename': 'document', 'Language': 'l1'})
    document_pairs = document_pairs[['document', 'l1']]
    sentence_pairs = get_sentence_pairs(document_pairs)

    return document_pairs, sentence_pairs


class BaseFeaturizer():
    '''
        This is the base class for a feature builder. A feature builder should take text
        data as input and return the numerical features which represent the text
        Implement the `featurize_all()` method if you intend to pass a series of samples
        Implement the `featurize()` method to convert a single string into a feature vector
    '''
    def __init__(self):
        pass

    def featurize_all(self, samples):
        '''
        The intended way this method should work:
            Samples is expected as a Pandas DataFrame
            The return type should be a Scipy csr sparse matrix
        :param samples:
        :return:
        '''
        raise NotImplementedError

    def featutize(self, text):
        '''
            This method should turn a single string into a Scipy csr sparse matrix
            This method will be used by `featurize_all()`
        :param text:
        :return:
        '''
        raise NotImplementedError


from langassist_model_engine.nlp_utils.features.builders import vocab_featurizer, hash_featurizer

def get_builder(builder_name):
    if(builder_name == "vocab"):
        return vocab_featurizer.VocabFeaturizer
    elif(builder_name == "hash"):
        return hash_featurizer.HashFeaturizer


from setuptools import find_packages, setup

setup(
    name="langassist_model_engine",
    version="0.1.1",
    url="https://langassist.rorybyrne.me",
    description="For building and maintaining the ML models in LangAssist",
    author="Rory Byrne",
    author_email="rory.byrne57@mail.dcu.ie",
    packages=find_packages(),
    python_requires='>=3.6',
    install_requires=['pandas',
                      'nltk',
                      'sklearn',
                      'numpy',
                      'scipy',
                      'boto3',
                      's3io']
)


import logging

import numpy as np
from sklearn import preprocessing
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import cross_val_score

from langassist_model_engine.native_lang_identification.model.base import BaseModel

logger = logging.getLogger(__name__)

class LogiReg(BaseModel):
    '''
    A wrapper class around Logistic Regression
    '''
    def __init__(self, feature_builder):
        BaseModel.__init__(self, feature_builder)
        self.le = preprocessing.LabelEncoder()
        self.model = LogisticRegression(penalty="l2", C=1.0)


    def fit_label_encoder(self, labels):
        logger.debug("Fitting labels to encoder...")
        self.le.fit(labels)

    def train(self, x_train, y_train):
        logger.info("Training...")
        self.model.fit(x_train, y_train)

    def test(self, x_test, y_test):
        logger.info("Testing...")
        preds = self.model.predict(x_test)

        scores = cross_val_score(self.model, x_test, y_test, cv=5)

        logger.info("Accuracy: %0.2f (+/- %0.2f)" % (scores.mean(), scores.std() * 2))

        # encoded_y = self.le.transform(y_test)
        # encoded_preds = self.le.transform(preds)
        #
        # for i, j in zip(preds, y_test):
        #     logger.debug("Expected %s, Guessed %s" % (j, i))
        #
        # logger.debug("Mean squared error: %.2f"
        #       % mean_squared_error(encoded_y, encoded_preds))
        # logger.debug("LogRegr .score(): %s" % self.model.score(x_test, y_test))
        # logger.info("F1 Score: %s" % f1_score(encoded_y, encoded_preds, average="micro"))

    def predict(self, predict_x):
        probs = self.model.predict_proba(predict_x)
        confidence = max(probs[0])
        res_index = np.argmax(probs[0])
        prediction = self.le.classes_[res_index]

        return (prediction, confidence)
